#include <stdexcept>
#include <iostream>
#include <sstream>
extern "C" {
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <libavutil/opt.h>
#include <libavfilter/buffersink.h>
#include <libavfilter/buffersrc.h>
}
#include "comparator.h"

Comparator::Comparator(const AVCodecContext *sampleCodecContext)
{
    const AVFilter *bufferImageA  = avfilter_get_by_name("buffer");
    const AVFilter *bufferImageB  = avfilter_get_by_name("buffer");
    const AVFilter *bufferSinkPsnr = avfilter_get_by_name("buffersink");
    const AVFilter *bufferSinkSsim = avfilter_get_by_name("buffersink");
    const AVFilter *psnrFilter  = avfilter_get_by_name("psnr");
    if(psnrFilter == nullptr)
        throw std::runtime_error("Cannot find psnr filter.");
    const AVFilter *ssimFilter  = avfilter_get_by_name("ssim");
    if(ssimFilter == nullptr)
        throw std::runtime_error("Cannot find ssim filter.");
    const AVFilter *splitFilterImageA  = avfilter_get_by_name("split");
    if(splitFilterImageA == nullptr)
        throw std::runtime_error("Cannot find split filter.");
    const AVFilter *splitFilterImageB  = avfilter_get_by_name("split");
    if(splitFilterImageB == nullptr)
        throw std::runtime_error("Cannot find split filter.");

    enum AVPixelFormat pix_fmts[] = { AV_PIX_FMT_NONE };

    filterGraph = avfilter_graph_alloc();
    if(!filterGraph)
        throw std::runtime_error("Cannot allocate filter graph.");

    std::stringstream arguments;
    arguments << "video_size=" << sampleCodecContext->width << "x" << sampleCodecContext->height <<
              ":pix_fmt=" << sampleCodecContext->pix_fmt <<
              ":time_base=" << 1 << "/" << sampleCodecContext->time_base.den <<
              ":pixel_aspect=" << sampleCodecContext->sample_aspect_ratio.num << "/" << sampleCodecContext->sample_aspect_ratio.den;
    if(avfilter_graph_create_filter(&bufferImageACtx, bufferImageA, "imageAIn", arguments.str().c_str(), nullptr, filterGraph) < 0)
        throw std::runtime_error("Cannot create filter graph imageAIn.");
    if(avfilter_graph_create_filter(&bufferImageBCtx, bufferImageB, "imageBIn", arguments.str().c_str(), nullptr, filterGraph) < 0)
		throw std::runtime_error("Cannot create filter graph imageBIn."); 

	if(avfilter_graph_create_filter(&bufferSinkPsnrCtx, bufferSinkPsnr, "outPsnr", nullptr, nullptr, filterGraph) < 0)
        throw std::runtime_error("Cannot create filter graph out for psnr.");
    if(avfilter_graph_create_filter(&bufferSinkSsimCtx, bufferSinkSsim, "outSsim", nullptr, nullptr, filterGraph) < 0)
        throw std::runtime_error("Cannot create filter graph out for psnr.");
    if(avfilter_graph_create_filter(&psnrFilterCtx, psnrFilter, "psnrFilter", nullptr, nullptr, filterGraph) < 0)
        throw std::runtime_error("Cannot create filter graph psnrFilter.");
    if(avfilter_graph_create_filter(&ssimFilterCtx, ssimFilter, "ssimFilter", nullptr, nullptr, filterGraph) < 0)
        throw std::runtime_error("Cannot create filter graph ssimFilter.");

    arguments.str("2");
    if(avfilter_graph_create_filter(&splitFilterImageACtx, splitFilterImageA, "splitFilterImageA", arguments.str().c_str(), nullptr, filterGraph) < 0)
        throw std::runtime_error("Cannot create filter graph splitFilterImageA.");
    if(avfilter_graph_create_filter(&splitFilterImageBCtx, splitFilterImageB, "splitFilterImageB", arguments.str().c_str(), nullptr, filterGraph) < 0)
        throw std::runtime_error("Cannot create filter graph splitFilterImageB.");

    if(av_opt_set_int_list(bufferSinkPsnrCtx, "pix_fmts", pix_fmts, AV_PIX_FMT_NONE, AV_OPT_SEARCH_CHILDREN) < 0)
        throw std::runtime_error("Cannot set pix fmt in graph.");
    if(av_opt_set_int_list(bufferSinkSsimCtx, "pix_fmts", pix_fmts, AV_PIX_FMT_NONE, AV_OPT_SEARCH_CHILDREN) < 0)
        throw std::runtime_error("Cannot set pix fmt in graph.");

    if(avfilter_link(bufferImageACtx, 0, splitFilterImageACtx, 0) != 0)
        throw std::runtime_error("Cannot link image A to split.");
    if(avfilter_link(bufferImageBCtx, 0, splitFilterImageBCtx, 0) != 0)
        throw std::runtime_error("Cannot link image B to split.");
        
    if(avfilter_link(splitFilterImageACtx, 0, ssimFilterCtx, 0) != 0)
        throw std::runtime_error("Cannot link split image A and ssim.");
    if(avfilter_link(splitFilterImageBCtx, 0, ssimFilterCtx, 1) != 0)
        throw std::runtime_error("Cannot link split image B and ssim.");
        
    if(avfilter_link(splitFilterImageACtx, 1, psnrFilterCtx, 0) != 0)
        throw std::runtime_error("Cannot link split image A and psnr.");
    if(avfilter_link(splitFilterImageBCtx, 1, psnrFilterCtx, 1) != 0)
        throw std::runtime_error("Cannot link split image B and psnr.");
  
    if(avfilter_link(psnrFilterCtx, 0, bufferSinkPsnrCtx, 0) != 0)
        throw std::runtime_error("Cannot link psnr and sink.");
    if(avfilter_link(ssimFilterCtx, 0, bufferSinkSsimCtx, 0) != 0)
        throw std::runtime_error("Cannot link ssim and sink.");
   
    if(avfilter_graph_config(filterGraph, nullptr) < 0)
        throw std::runtime_error("Cannot configure filter graph.");
}

float Comparator::getMetric(AVFilterContext *filterContext, std::string dataName)
{
	if(av_buffersink_get_frame(filterContext, resultFrame) < 0)
		throw std::runtime_error("Cannot recieve frame from the graph.");
    auto metricData = av_dict_get(resultFrame->metadata, dataName.c_str(), nullptr, 0)->value;
    float metricValue = std::stof(metricData);
    av_frame_unref(resultFrame);
    return metricValue;
}

void Comparator::printMetrics()
{
	std::cout << "PSNR: " << getMetric(bufferSinkPsnrCtx, "lavfi.psnr.psnr_avg") << std::endl;
    std::cout << "SSIM: " << getMetric(bufferSinkSsimCtx, "lavfi.ssim.All") << std::endl;
}

void Comparator::pushImageA(AVFrame *frame)
{
    if(av_buffersrc_add_frame_flags(bufferImageACtx, frame, AV_BUFFERSRC_FLAG_KEEP_REF) < 0)
        throw std::runtime_error("Cannot add image A frame to the graph.");    
}

void Comparator::pushImageB(AVFrame *frame)
{ 
    if(av_buffersrc_add_frame_flags(bufferImageBCtx, frame, AV_BUFFERSRC_FLAG_KEEP_REF) < 0)
        throw std::runtime_error("Cannot add image B frame to the graph.");
}

Comparator::~Comparator()
{
	av_frame_free(&resultFrame);
	avfilter_graph_free(&filterGraph);
}
